/*
A GLUT program using OpenGL, GLUI, and GLSL.

Orion Sky Lawlor, olawlor@acm.org, 2007/01/18 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

/* Dr. Lawlor's GLSL utilities */
#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"


/* Variables updated by the GUI */
float distance=2.0; /* origin-to-camera distance */
vec3 cameraloc; /* location of the camera */
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

float brightness=1.0, shininess=1.0;
int draw_spheres=0; /* if 1, draw bounding spheres */
int draw_lines=1; /* if 1, draw outlines instead of filling quads */
float error_threshold=20.0; /* onscreen pixel size threshold */

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 modelMat(1.0), viewMat(1.0);
mat4 clipMat(1.0);

/// Scale by this offset
inline void glScalefv(const float *s) {
	glScalef(s[0],s[1],s[2]);
}
/// Translate by this offset
inline void glTranslatefv(const float *s) {
	glTranslatef(s[0],s[1],s[2]);
}
/** Transpose this mat4.  If the mat4 was a rotation matrix, 
  this actually inverts the matrix. */
mat4 transpose(const mat4 &m) {
	return mat4(
		m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}



// Return the terrain height at this point
float get_height(const vec3 &v) {
	return 0.04*sin(10.0*v.x)+0.05*cos(11.0*v.y);
}
// Return the terrain normal at this point
//  normal is (dh/dx, dh/dy, 1)
//  FIXME: store the normals (or even all lighting) in a texture; don't recompute every frame!
vec3 get_normal(const vec3 &v) {
	return vec3(0.04*10.0*cos(10.0*v.x),-0.05*11.0*sin(11.0*v.y),1.0);
}

/* This class looks at a bounding volume and decides if it should be clipped */
class clipper {
public:
	/**
	  Definition: a point p is inside the clipping plane if 
	  	dot(n,p)+c >= 0
	*/
	vec3 n; float c;
	
	/**
	  Rationale: a bounding volume is outside the clipping plane if all of its points are.
	*/
	bool is_outside(const vec3 &center,float radius) const {
		return (dot(n,center)+c+radius)<=0.0;
	}
};
clipper clip;


/* Here's one node of the terrain tree */
class terrain_treenode {
public:
	enum {nkids=4};
	terrain_treenode *kids[nkids];
	
	vec3 center; /* Center of node */
	float radius; /* Size of node */
	
	terrain_treenode(vec3 center_,float radius_) 
		:center(center_), radius(radius_)
	{
		center.z=get_height(center);
		if (radius<=0.001) {
			for (int i=0;i<nkids;i++) kids[i]=0;
			return;
		}
		float r=radius*0.5;
		kids[0]=new terrain_treenode(center+vec3(-r,+r,0),r);
		kids[1]=new terrain_treenode(center+vec3(+r,+r,0),r);
		kids[2]=new terrain_treenode(center+vec3(-r,-r,0),r);
		kids[3]=new terrain_treenode(center+vec3(+r,-r,0),r);
	}
	
	// Recursively draw this treenode 
	void draw(void);
	// Draw this treenode as a leaf
	void draw_leaf(void);
	// Draw this vertex, with appropriate normal
	void draw_vertex(const vec3 &v);
	// Draw this treenode's bounding sphere
	void draw_sphere(void);
};

// This is the central function of this class: recursively draw this tree node.
void terrain_treenode::draw(void)
{
	// If we're offscreen, forget it.
	if (clip.is_outside(center,radius)) return;
	
	// If we don't have kids, we don't have to draw them!
	if (kids[0]==0) {draw_leaf(); return;}
	
	// Check if the kids are too far away to be drawn
	float onscreen_size=radius*win_w/length(cameraloc-center);
	if (onscreen_size<error_threshold)
	{ /* don't draw children, draw us instead */
		draw_leaf();
		return;
	}
	
	// Otherwise just draw all our kids
	for (int i=0;i<nkids;i++) 
	{
		kids[i]->draw();
	}
}

// Draw this treenode as a leaf
void terrain_treenode::draw_leaf(void)
{
	if (draw_spheres) draw_sphere();
	glColor4f(1,1,1,1);
	if (draw_lines) {
		glBegin(GL_LINE_LOOP);
	} else {
		glBegin(GL_QUADS);
	}
	draw_vertex(center+vec3(-radius,+radius,0));
	draw_vertex(center+vec3(+radius,+radius,0));
	draw_vertex(center+vec3(+radius,-radius,0));
	draw_vertex(center+vec3(-radius,-radius,0));
	glEnd();
}

// Draw this vertex, with appropriate normal
void terrain_treenode::draw_vertex(const vec3 &v)
{
	float h=get_height(v); // height
	glNormal3fv(get_normal(v));
	glVertex3f(v.x,v.y, h);
}

// Draw this treenode's bounding sphere
void terrain_treenode::draw_sphere(void)
{
	glPushMatrix();
	glColor4f(1,0,0,0.3);
	glTranslatefv(center);
	glutSolidSphere(radius*1.42,8,6);
	glPopMatrix();
}

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	// Clear the color and depth buffers
	glClearColor(0.2,0.4,0.8,0.0); /* sky blue */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	/* Depth testing */
	glEnable(GL_DEPTH_TEST);
	
	/* Alpha blending */
	glEnable(GL_BLEND);
	           /*  src color  + dest color  */
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	// Load all needed matrices into OpenGL
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); /*<- clean out any old leftover matrices */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		win_w/(float)win_h, /* viewport's aspect ratio: always width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	glTranslatef(0,0,-distance); /* push world origin down the *unrotated* Z axis */
	glMultMatrixf(&viewMat[0][0]); /* camera orientation: could also go under GL_MODELVIEW */
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	/* Set up programmable shader. */
	static oglProgramObject *p=oglProgramFromFiles("vertex.txt","fragment.txt");
	p->begin();
	p->set("brightness",brightness); /* full diffuse intensity */
	
	/* World location of camera: just back out all the matrix transforms */
	cameraloc=vec3(transpose(viewMat)*vec4(0,0,+distance,1));
	p->set("cameraloc",cameraloc);
	
	/* Set up shading */
	p->set("shininess",0.0f); /* no specular intensity */
	p->set("shininess",shininess); /* full specular intensity */
	
	// Set up clipping
	clip.n=normalize(vec3(1,1,0)); clip.c=+2.0;
	
	// Draw the terrain:
	static terrain_treenode *tn=new terrain_treenode(vec3(0,0,0),2.0);
	tn->draw();
	
	p->end();
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w; win_h=h;
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)(win_h));
}

int main(int argc, char** argv)
{
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Lighting Demo");

	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();
	
	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	
	distance=2;
	/* Set up view matrix so it's tilted forward a bit */
	vec3 up=normalize(vec3(0,1,0.2));
	viewMat[1]=vec4(up,0.0);
	viewMat[2]=vec4(vec3(up.x,-up.z,up.y),0.0);

	if (1) {
		// Make some sliders (this is the only GLUI code you'll need!)
		GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
			GLUI_SUBWINDOW_TOP);
		/* This causes the main window to be redrawn when a GLUI control changes */
		glui->set_main_gfx_window( main_window );

		(new GLUI_Spinner( glui, "Distance:", &distance))->set_float_limits(0.0, 100.0);
		(new GLUI_Spinner( glui, "Error:", &error_threshold))->set_float_limits(0.0, 100.0);
		
		new GLUI_Checkbox(glui,"Bounds",&draw_spheres);
		new GLUI_Checkbox(glui,"Outlines",&draw_lines);
		
		new GLUI_Column(glui);
		//new GLUI_Rotation(glui,"Model:",&modelMat[0][0]);
		//new GLUI_Column(glui,false);
		new GLUI_Rotation(glui,"View:",&viewMat[0][0]);
	}
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
